#
#    Copyright (C) 2015  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.W
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import re
from DateTime import DateTime
from DateTime.DateTime import _MONTH_LEN

valid_char_number = re.compile(r'''^\d+$''')
valid_char_list = re.compile(r'''^\d+(-\d+)*(/\d+)*$''')
numbers = re.compile(r'''(\d+)''')

def parse_spec(specification, min, max):
    """
    crontab spec parser/validater
    """
    specification.replace(' ', '')
    if specification == '*': return 1
    for element in specification.split(','):
        if not valid_char_number.match(element):
            if not valid_char_list.match(element):
                return 0
            else:
                begin,end=element.split('-')
                if end.find('/') != -1:
                    end,increment=end.split('/')
                    if not valid_char_number.match(increment):
                        return 0
                if not valid_char_number.match(begin) and not valid_char_number.match(end):
                    return 0
        else:
            if int(element) < min or int(element) > max:
                return 0
    return 1

def expand_spec(specification):
    """
    take  a-b/c and produce a comma-separated list of expanded numbers
    this is a prereq for plone4cron
    """
    specification = specification.replace(' ', '')

    if specification.find(',') != -1:
        parts = specification.split(',')
        return ','.join(map(lambda x: expand_spec(x), parts))

    mod = 0

    if specification.find('/') != -1:
        specification, mod = specification.split('/')
        mod = int(mod)

    if specification.find('-') != -1:
        smin, smax = specification.split('-')
        smin = int(smin)
        smax = int(smax)
        values = range(smin, smax + 1)
    else:
        values = [specification]

    if mod:
        tmp = [smin]
        while smin < smax:
            smin += mod
            if smin <= smax:
                tmp.append(smin)
        values = tmp

    return ','.join(map(lambda x: str(x), values))
        

def get_first_time_for_spec(specification, min):
    """
    return the first valid number from the spec
    """
    if specification == '*':
        return min
    match = numbers.search(specification)
    return match.group(0)
    
def get_time_for_spec(specification, value, min, max):
    """
    specification is a crontab range, value is what you currently have, min
    and max are the boundaries to this time component

    returns a tuple of the next valid value and a flag indicating if
    it had to wrap to discover this - an indication to increment the next
    time component ;)
    """
    if specification == '*': 
        expansion = range(min, max + 1)
    else:
        expansion = []
        for element in specification.split(','):
            if valid_char_number.match(element):
                expansion.append(int(element))
            else:
                begin,end=element.split('-')
                begin = int(begin)
                if end.find('/') != -1:
                    end,increment=end.split('/')
                    end = int(end)
                    increment = int(increment)
                    index = begin
                    while index <= end:
                        expansion.append(index)
                        index += increment
                else:
                    end = int(end)
                    expansion.extend( range( begin, end + 1 ) )

    expansion.sort()

    for element in expansion:
        if element >= value:
            return (element, 0)
    return (expansion[0], 1)

def next_day(tm, tz):
    """
    returns the beginning of the next day, allowing for month/year rollover
    """
    m = tm.month()
    y = tm.year()
    month_len = _MONTH_LEN[(y%4==0 and (y%100!=0 or y%400==0))][m]
    d = tm.day()
    if d < month_len:
        return DateTime('%s/%s/%0.2i 00:00 %s' % (y, tm.mm(), d+1, tz))
    else:
        return next_month(tm, tz)

def next_month(tm, tz):
    """
    returns the beginning of the next month, allowing for year rollover
    """
    month = tm.month()
    if month == 12:
        return DateTime('%i/01/01 00:00 %s' % (tm.year() + 1, tz))
    else:
        return DateTime('%s/%0.2i/01 00:00 %s' % (tm.year(), month + 1, tz))





#
#    Copyright (C) 2011  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import AccessControl
from AccessControl.Permissions import change_configuration
from Products.ZScheduler.config import TOOLNAME
from Products.ZScheduler.interfaces import ISchedulerTool
from OFS.SimpleItem import SimpleItem
from Products.CMFCore.ActionProviderBase import ActionProviderBase
from Products.CMFCore.utils import registerToolInterface, UniqueObject
from zope.interface import implements

class SchedulerTool(UniqueObject, ActionProviderBase, SimpleItem):
    """
    A tool to manage scheduled events in your portal
    """
    meta_type = portal_type = 'SchedulerTool'

    implements(ISchedulerTool)

    id = TOOLNAME
    title = 'Scheduled Events Tool'
    _actions = ()

    __ac_permissions__ = SimpleItem.__ac_permissions__ + (
        (change_configuration, ('queueValues', 'timezone', 
                                'schedule', 'unschedule')),
        ) +  ActionProviderBase.__ac_permissions__

    manage_options = ActionProviderBase.manage_options + \
        SimpleItem.manage_options


    def __init__(self, id=TOOLNAME):
        pass

    def queueValues(self):
        """
        """
        tool = self.getPhysicalRoot().ZSchedulerTool
        query = {'getPhysicalPath': '/'.join(self.aq_parent.getPhysicalPath()) }
        return map(lambda x: x.getObject(), tool.searchResults(**query))

    def timezone(self):
        """
        """
        tool = self.getPhysicalRoot().ZSchedulerTool
        return tool.timezone

    def schedule(self, urls):
        """
        start scheduling the indicated urls
        """
        my_url = '/'.join(self.aq_parent.getPhysicalPath())
        tool = self.getPhysicalRoot().ZSchedulerTool
        tool.manage_schedule(filter(lambda x: x.startswith(my_url), urls))
        
    def unschedule(self, urls):
        """
        stop scheduling the indicated urls
        """
        my_url = '/'.join(self.aq_parent.getPhysicalPath())
        tool = self.getPhysicalRoot().ZSchedulerTool
        tool.manage_unschedule(filter(lambda x: x.startswith(my_url), urls))

AccessControl.class_init.InitializeClass(SchedulerTool)


registerToolInterface(TOOLNAME, ISchedulerTool)



#
#    Copyright (C) 2004-2011  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from threading import Thread
import AccessControl, transaction
from Products.ZScheduler.timers.ZopeSingleThreaded.ZopeSingleThreaded import ZopeSingleThreaded

class ZopeMultiThreaded(ZopeSingleThreaded):
    """
    Timer thread kicks off job(s) in their own detached thread.

    This has the benefit of most accurately maintaining job start times
    as the timer is not delayed in the task of dispatch by performing
    potentially long-running schedule events.
    """
    meta_type = 'ZopeMultiThreaded'

    def _dispatch(self, event):
        worker = Thread(None, self._invokeFromPath, ['/'.join(event.getPhysicalPath())])
        worker.setName(event.getId())
        worker.setDaemon(1)                # detach thread ...
        worker.start()
        
    def _invokeFromPath(self, event_path):
        # Set up the "application" object that automagically opens connections

        # TODO - fix all this up!!!
        conn = Globals.DB.open()
        app = self._getContext(conn.root()['Application'])

        app.unrestrictedTraverse(event_path).manage_invokeEvent()

        transaction.get().commit()
        conn.close()

AccessControl.class_init.InitializeClass(ZopeMultiThreaded)



#    Copyright (C) 2004-2011  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import AccessControl
from AccessControl import ClassSecurityInfo
from Products.ZCatalog.ZCatalog import ZCatalog

from interfaces.IQueue import IQueue
from threading import Lock

from Acquisition import aq_base
from ZScheduleEvent import ZScheduleEvent
from Exceptions import ZSchedulingException

from zope.interface import implements

class ZQueue(ZCatalog):
    """
    This is a thread-safe queue of scheduled events and we're getting loads for free as
    we've set up a Date index on it's key ...
    """
    implements(IQueue)

    meta_type = 'ZQueue'
    title = 'ZQueue'
    _lock = Lock()
    manage_options = ZCatalog.manage_options[1:]
    manage_main = ZCatalog.manage_catalogView
    _security = ClassSecurityInfo()
    
    def all_meta_types(self):
        """ contain nothing """
        return []
    
    def __init__(self, id):
        ZCatalog.__init__(self, id)
        # TODO - implement status
        self.addIndex('time', 'DateIndex')
        self.addIndex('getPhysicalPath', 'PathIndex')
        self.addIndex('active', 'BooleanIndex')
        for name in ('time', 'meta_type',):
            self.addColumn(name)

    def pop(self, obj):
        """
        remove obj from catalog (actually just recatalog to force status change)
        """
        url = '/'.join(obj.getPhysicalPath())
        try:
            self._lock.acquire()
            self.catalog_object(obj, url)
        finally:
            self._lock.release()
            
    push = pop

    def get(self, url):
        try:
            self._lock.acquire()
            return self.getobject( self.getrid(url) )
        finally:
            self._lock.release()
            
    def delete(self, url):
	try:
            self._lock.acquire()
            self.uncatalog_object(url)
        finally:
            self._lock.release()
	
    def reload(self, REQUEST=None):
        """
        search ZODB for all ZScheduleEvents and place them in the queue
        """
        root = self.getPhysicalRoot()
        try:
            self._lock.acquire()
            #
            # hmmm - we'd really like to do this but this doesn't work
            # for (i) subclassed ZScheduleEvent's; or (ii) stuff not in
            # ObjectManager containment relationships :(
            #
            #    self.ZopeFindAndApply(root,
            #                          obj_metatypes=['ZScheduleEvent'],
            #                          search_sub=1,
            #                          apply_func=self.catalog_object,
            #                          apply_path='/'.join(root.getPhysicalPath()))
            self._reschedule(root)
        finally:
            self._lock.release()

        if REQUEST:
	    return self.manage_main(self, REQUEST)

    def getNext(self):
        try:
            self._lock.acquire()
            try:
                return self()[0].getObject()
            except:
                return None
        finally:
            self._lock.release()

    def _reschedule(self, folder):
        """
        tail-recursive ZODB rescheduler

        this function is designed to work with (i) subclasses of ZScheduleEvent;
        (ii) folder-like object attributes that aren't necessarily included in
        the _objects member ;)
        """
	try:
            objs = folder.objectValues()
	except Exception, e:
	    raise ZSchedulingException, '%s\n%s' % (folder.absolute_url(), str(e))

        # we are calling the ZCatalog stuff directly to avoid deadlock
        for e in filter( lambda x: isinstance(x, ZScheduleEvent), objs):
	    try:
                e.time = e.nextEventTime()
	    except:
		raise ZSchedulingException, e
            self.catalog_object(e, '/'.join(e.getPhysicalPath()))

        # recurse through all sub-folders ...
        for sf in filter(lambda x: getattr(aq_base(x), 'objectValues', None), objs):
            try:
                self._reschedule(sf)
            except:
                # hmmm - non-iterable objects etc ...
                continue

    def _repair(self):
        try:
            self.addIndex('getPhysicalPath', 'PathIndex')
        except:
            pass
        try:
            self.addIndex('active', 'BooleanIndex')
        except:
            pass
        try:
            self.addColumn('meta_type')
        except:
            pass
        self.manage_reindexIndex(['time', 'getPhysicalPath', 'active'])

AccessControl.class_init.InitializeClass( ZQueue )



#
# Copyright 2012 Corporation of Balclutha (http://www.balclutha.org)
# 
#                All Rights Reserved
# 
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
#
# Corporation of Balclutha DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS, IN NO EVENT SHALL Corporation of Balclutha BE LIABLE FOR
# ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE. 
#
from Acquisition import aq_base, aq_inner
from Products.Five.browser import BrowserView

from Products.ZScheduler.ZScheduler import ZScheduler



class ZSchedulerTool(BrowserView):
    """
    Creates a Zenoss installation
    """

    def createTool(self):
        """
        creates zport, DMD in install root
        """
        context = aq_inner(self.context)

	context._setObject('ZSchedulerTool', ZScheduler())	

        self.request.set('manage_tabs_message', 'created ZSchedulerTool')
        self.request.RESPONSE.redirect('/manage_main')




#
#    Copyright (C) 2004-2015  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import atexit, logging, transaction
import ZScheduler, ZScheduleEvent, FSZScheduleEvent
from App.ImageFile import ImageFile
from AccessControl import ModuleSecurityInfo
from AccessControl.Permissions import view
from Permissions import add_scheduler_events
from Acquisition import aq_base
import ZODB

from config import *

# optional plone stuff (probably no longer optional)
try:
    import tool
    from Products.CMFCore.DirectoryView import registerDirectory
    from Products.CMFCore import utils
    from Products.CMFCore.permissions import AddPortalContent
    registerDirectory(SKINS_DIR, GLOBALS)
    DO_PLONE=1
except:
    DO_PLONE=0

LOG = logging.getLogger('ZScheduler')
 

def initialize(context):
    """
    hook into control panel
    """
    context.registerClass(ZScheduleEvent.ZScheduleEvent,
                          constructors = (ZScheduleEvent.manage_addZScheduleEventForm,
                                          ZScheduleEvent.manage_addZScheduleEvent),
                          icon='www/event.gif',
                          permission=add_scheduler_events
                          )
    context.registerHelp()

    if DO_PLONE:
        utils.registerIcon(FSZScheduleEvent.FSZScheduleEvent, 'www/fsevent.gif', GLOBALS)

        utils.ToolInit('ZScheduler Tool',
                       tools=(tool.SchedulerTool,),
                       icon='tool.gif').initialize(context)


    if REGISTER_SCHEDULER:
        global scheduler
        try:
            from Zope2 import bobo_application
        except ImportError:
            bobo_application = None
        if bobo_application is not None:
            app = bobo_application()
        else:
            app = context._ProductContext__app

        if app._getOb('ZSchedulerTool', None) is None:
            app._setObject('ZSchedulerTool', ZScheduler.ZScheduler())
            try:
                #transaction.get().savepoint(optimistic=True)
                transaction.get().commit()
            except ZODB.POSException.StorageTransactionError:
                pass

        scheduler = app._getOb('ZSchedulerTool', None)
        if scheduler is not None:
            try:
                # don't want to hang the server if an exception is raised ...
                scheduler._start()
            except Exception, e:
                LOG.error('scheduler worker start failed', exc_info=True)
        
            # clean up stuff on exit ...
            atexit.register(scheduler._stop)
        else:
            LOG.error('eek - no scheduler created/found')

misc_ = {}
for icon in ('scheduler', 'logrec', 'logger', 'timer'):
    misc_[icon] = ImageFile('www/%s.gif' % icon, GLOBALS)

from AccessControl import ModuleSecurityInfo
ModuleSecurityInfo('Products').declarePublic('ZScheduler')
ModuleSecurityInfo('Products.ZScheduler').declarePublic('ZScheduleEvent')
ModuleSecurityInfo('Products.ZScheduler.ZScheduleEvent').declarePublic('parse_spec')





#
# Skeleton ZopeTestCase
#
from Testing import ZopeTestCase

class TestDummy(ZopeTestCase.ZopeTestCase):

    def setUp(self):
	# runalltests gets confused ...
	pass

    def testImport(self):
        # check that it loads OK - we don't get to see normal import errors :(
        from Products.ZScheduler.timers.Dummy import Dummy


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestDummy))
    return suite















#
#    Copyright (C) 2004-2012  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import subprocess, logging, AccessControl
from AccessControl.Permissions import view, change_configuration
from Products.ZScheduler.interfaces.ITimer import ITimer
from Products.ZScheduler.timers.Base import ThreadedTimer
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

log = logging.getLogger('ZScheduler.Crontab')

class Crontab(ThreadedTimer):
    """
    A crontab dispatching mechanism
    """
    meta_type = 'Crontab'
    __implements__ = (ITimer,)

    __ac_permissions__ = ThreadedTimer.__ac_permissions__ + (
        (change_configuration, ('manage_load', 'manage_unload')),
        (view, ('index_html',)),
        )

    property_extensible_schema__ = 0
    _properties = ThreadedTimer._properties + (
        {'id':'command',   'mode':'w', 'type':'string'},
        )

    manage_options = (
        {'label':'Properties', 'action':'manage_propertiesForm',
         'help':('ZScheduler', 'crontab.stx')},
        {'label':'View', 'action':''},
        ) + ThreadedTimer.manage_options[1:]

    manage_main = ThreadedTimer.manage_propertiesForm
    
    def __init__(self):
        ThreadedTimer.__init__(self)
        self.command='wget -q --tries=1 --user=admin --password=secret --auth-no-challenge -O /dev/null'
        
    index_html = PageTemplateFile('zpt/crontab', globals())

    def _load(self):
        """
        physically install our crontab
        """
        crontab = self.index_html()

        log.info(crontab)

        try:
            pipe = subprocess.Popen(['crontab', '-'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        except OSError, e:
            log.error(str(e))
            return

        output = pipe.communicate(crontab)

        if output:
            log.error('load error: %s' % str(output))

    def manage_load(self, REQUEST=None):
        """
        """
        self._load()
        if REQUEST:
            REQUEST.set('manage_tabs_message', 'Load crontab')
            return self.manage_main(self, REQUEST)
            
    def _unload(self):
        """
        remove our system crontab
        """
        try:
            pipe = subprocess.Popen(['crontab', '-r'], stdout=subprocess.PIPE, stderr=subprocess.PIPE )
        except OSError, e:
            log.error(e)
            return

        output,errors = pipe.communicate()

        if output:
            log.error(output)

    def manage_unload(self, REQUEST=None):
        """
        """
        self._unload()
        if REQUEST:
            REQUEST.set('manage_tabs_message', 'Unloaded crontab')
            return self.manage_main(self, REQUEST)
            

    def manage_editProperties(self, command, is_active=False, REQUEST=None):
        """
        change all properties, then tell ZScheduler to reload crontab ...
        """
        old_is_active = self.is_active
        old_command = self.command

        self._updateProperty('is_active', is_active)
        self._updateProperty('command', command)

        if old_command != command:
            if old_is_active and is_active:
                self.manage_restart()

        if is_active:
            if not old_is_active:
                self._start()
            else:
                self._stop()
        
        # tell cron timer to reload - f**k knows why it doesnt' see the latest timer props!
        self.aq_parent.semaphore.set()
        if REQUEST:
            REQUEST.set('manage_tabs_message', 'Properties Updated')
            return self.manage_propertiesForm(self, REQUEST)


    def isActive(self):
        """
        verify the crontab is installed
        """
        try:
            pipe = subprocess.Popen(['crontab', '-l'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except OSError, e:
            return False

        output, errors = pipe.communicate()
        if output and len(output) > 25:
            return True
        return False
            
AccessControl.class_init.InitializeClass(Crontab)



#
#    Copyright (C) 2005-2011  Corporation of Balclutha. All rights Reserved.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os

from Testing import ZopeTestCase
ZopeTestCase.installProduct('ZScheduler')

try:
    __file__
except NameError:
    # Test was called directly, so no __file__ global exists.
    _prefix = os.path.abspath(curdir)
else:
    # Test was called by another test.
    _prefix = os.path.abspath(os.path.dirname(__file__))


class TestFSScheduleEvent(ZopeTestCase.ZopeTestCase):

    def _makeOne( self, id, filename='event.sched'):

        from Products.ZScheduler.FSZScheduleEvent import FSZScheduleEvent

        return FSZScheduleEvent( id, os.path.join(_prefix, filename) )

    def _extractFile( self ):

        path = os.path.join(_prefix, 'event.sched')
        f = open( path, 'rb' )
        try:
            data = f.read()
        finally:
            f.close()

        return path, data

    def test_ctor( self ):

        path, ref = self._extractFile()

        event = self._makeOne( 'test_file')

        self.assertEqual(event.day_of_month, '*' )
        self.assertEqual(event.day_of_week, '*' )
        self.assertEqual(event.hour, '1' )
        self.assertEqual(event.minute, '1' )
        self.assertEqual(event.month, '3,6,9,12' )
        self.assertEqual(event.tz, 'EADT' )


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestFSScheduleEvent))
    return suite













#
#    Copyright (C) 2015  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from Acquisition import aq_inner
from zope.interface import implements, Invalid
from zope import schema
from z3c.form import button
from plone.supermodel import model

from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from Products.CMFPlone import PloneMessageFactory as _
from Products.statusmessages.interfaces import IStatusMessage
from plone.directives import form

from ..utils import parse_spec

def minuteConstraint(value):
    if not parse_spec(value, 0, 59):
        raise Invalid(_(u"Invalid minute range"))
    return True

def hourConstraint(value):
    if not parse_spec(value, 0, 23):
        raise Invalid(_(u"Invalid hour range"))
    return True

def monthConstraint(value):
    if not parse_spec(value, 1, 12):
        raise Invalid(_(u"Invalid month range"))
    return True

def dowConstraint(value):
    if not parse_spec(value, 0, 7):
        raise Invalid(_(u"Invalid day of week range"))
    return True

def domConstraint(value):
    if not parse_spec(value, 1, 31):
        raise Invalid(_(u"Invalid day of month range"))
    return True


class IPropertySchema(model.Schema):
    """
    Registration properties
    """

    minute = schema.ASCIILine(
        title=_(u'label_minute', default=u'Minute'),
         description=_(u'help_minute',
                      default=u"Which minute (0-59)."),
        required = True,
        constraint = minuteConstraint,
        default = '0',
        )

    hour = schema.ASCIILine(
        title=_(u'label_hour', default=u'Hour'),
         description=_(u'help_hour',
                       default=u"Which hour (0-23)."),
        required = True,
        constraint = hourConstraint,
        default = '0'
        )

    month = schema.ASCIILine(
        title=_(u'label_Month', default=u'Month'),
         description=_(u'help_month',
                       default=u"Which day (0-12)."),
        required = True,
        constraint = monthConstraint,
        default = '*'
        )

    day_of_month = schema.ASCIILine(
        title=_(u'label_day_of_month', default=u'Day of Month'),
         description=_(u'help_day_of_month',
                       default=u"Which day (1-31)."),
        required = True,
        constraint = domConstraint,
        default = '*'
        )

    day_of_week = schema.ASCIILine(
        title=_(u'label_day_of_week', default=u'Day of Week'),
         description=_(u'help_day_of_week',
                       default=u"Which day of week (0-7). 0 and 7 are Sunday"),
        required = True,
        constraint = dowConstraint,
        default = '*'
        )

    tz = schema.Choice(
        title=_(u'label_timezone', default=u'Timezone'),
        description=_(u'help_timezone',
                      default=u"Which timezone."),
        required = True,
        vocabulary='plone.app.vocabularies.Timezones'
        )

    active = schema.Bool(
        title=_(u'label_active', default=u'Active'),
         description=_(u'help_active',
                      default=u"Is this schedule currently active (invoked)."),
        required = True,
        default = False,
        )




class EditForm(form.SchemaForm):
    schema = IPropertySchema

    label = _(u"Event Schedule")
    description = _(u"Scheduling of background(s) on this item.")

    def update(self):
        # disable Plone's editable border
        #self.request.set('disable_border', True)

        # call the base class version - this is very important!
        super(EditForm, self).update()

    def updateActions(self):
        super(EditForm, self).updateActions()
        self.actions["edit"].addClass("context")
        self.actions["cancel"].addClass("standalone")

    @button.buttonAndHandler(_(u'Edit'))
    def handleEdit(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        try:
            self.context.manage_editSchedule(data['tz'], 
                                             data['minute'],
                                             data['hour'], 
                                             data['month'], 
                                             data['day_of_month'], 
                                             data['day_of_week'], 
                                             data['active'])
        except Exception, e:
            IStatusMessage(self.request).addStatusMessage(str(e),"error")
            self.status = self.formErrorsMessage
            return

        # Redirect back to the front page with a status message
        IStatusMessage(self.request).addStatusMessage(_(u"Scheduler Properties Edited"),"info")
        contextURL = self.context.absolute_url()
        self.request.response.redirect(contextURL)

    @button.buttonAndHandler(_(u"Cancel"))
    def handleCancel(self, action):
        """User cancelled. Redirect back to the front page.
        """
        contextURL = self.context.absolute_url()
        self.request.response.redirect(contextURL)


class ViewForm(EditForm):

    mode = 'display'



#
#    Copyright (C) 2004-2015  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import AccessControl, re, string
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from DateTime import DateTime, Timezones
from DateTime.DateTime import _MONTH_LEN
from Acquisition import aq_base
from Log import Logger, LogSupport
from AccessControl import ClassSecurityInfo
from AccessControl.Permissions import view_management_screens, change_configuration, view
from Permissions import execute_scheduler_events
from interfaces import IScheduleEvent
from zope.interface import implements
from browser.zscheduleevent import IPropertySchema

from utils import *

class ZScheduleEvent (LogSupport):
    """
    A schedule event, this wraps a callable with scheduling information

    Note that an event IS NOT scheduled on construction - it must be
    explicitly activated via editing it's status or calling manage_schedule
    """
    meta_type = portal_type = 'ZScheduleEvent'

    implements(IScheduleEvent, IPropertySchema)

    __ac_permissions__ = (
        (view, ('cronTab', 'expandedCronTab', 'active')),
        (view_management_screens, ('manage_scheduleForm', 'nextEventTime', 'event',
                                   'test_schedule')),
        (change_configuration, ('manage_editSchedule', 'manage_schedule',
                                'manage_unschedule',)),
        (execute_scheduler_events, ('manage_invokeEvent',)),
        ) + LogSupport.__ac_permissions__
    
    manage_options = (
        {'label':'Schedule', 'action':'manage_main', 'help':('ZScheduler', 'event.stx') },
        {'label':'Trigger', 'action':'manage_invokeEvent'},
        ) + LogSupport.manage_options
    
    last_executed = None
    _active = False

    manage_main = manage_scheduleForm = PageTemplateFile('zpt/event', globals())

    def __init__(self, id, title, callable, minute='*', hour='*', month='*', day_of_month='*', day_of_week='*', tz='UTC'):
        self.id = id
        self.title = title
        self.callable_id = callable
        self.minute = minute
        self.hour = hour
        self.month = month
        self.day_of_month = day_of_month
        self.day_of_week = day_of_week
        self.tz = tz
        self.time = DateTime(0).toZone(self.tz)
        LogSupport.__init__(self)

    def nextEventTime(self, time=None):
        """
        return next executable time after given time
        """
	# hmmm - tz is a new parameter ...
	if not getattr(self, 'tz', None):
	    self.tz = 'UTC'
        
        if time is None:
            time = DateTime(self.tz)

        # add one minute ...
        new_minute,increment_hr = get_time_for_spec('0-59', time.minute()+1, 0, 59)
        hour = int(time.hour())
        if increment_hr:
            hour += 1
        if hour == 24:
            my_time =  DateTime('%s/%s/%s 00:%0.2i %s' % (
                time.year(), time.mm(), time.dd(), new_minute, time.timezone() ) ) + 1
        else:
            my_time = DateTime('%s/%s/%s %0.2i:%0.2i %s' % (
                time.year(), time.mm(), time.dd(), hour, new_minute, time.timezone() ) )
                                                                                           
        while not self._matches(my_time):
            my_time = self._doMonth(self._doDay(self._doHour(self._doMinute(my_time))))

        return my_time

    def event(self):
        """
        return the actual event object/method

        the event object's security permissions control access
        """
        try:
            return self.aq_parent.restrictedTraverse(self.callable_id)
        except KeyError:
            # hmmm - must be a method ...
            return getattr(self, self.callable_id)

    @property
    def active(self):
        """
        active status flag
        """
        #return getattr(aq_base(self), '_active', False)
        return self._active
    
    def timezones(self):
        """
        return known timezones
        """
        return Timezones()
    
    def manage_beforeDelete(self, item, container):
        """
        delete event from scheduler, also must delete itself if the callable is deleted ...
        """
        self.getPhysicalRoot().ZSchedulerTool.unschedule(self)
        # this bit doesn't work ...
        if item.getId() == self.callable_id:
            container._delObject(self.getId())
                                 
    def manage_editSchedule(self, tz, minute, hour, month, day_of_month,
                            day_of_week, active=False, REQUEST=None):
        """
        update the schedule info ...
        """
        errors = []
        if not parse_spec(minute, 0, 59): errors.append( "Invalid Minute" )
        if not parse_spec(hour, 0, 23): errors.append( "Invalid Hour" )
        if not parse_spec(month, 1, 12): errors.append( "Invalid Month" )
        if not parse_spec(day_of_week, 0, 7): errors.append( "Invalid Day of Week" )
        if not parse_spec(day_of_month, 1, 31): errors.append( "Invalid Day of Month" )
        if active and (minute == '*' and hour == '*' and month == '*' and \
                 day_of_week == '*' and day_of_month == '*'):
            errors.append("Cannot assign task to run every minute forever!")
        if errors:
            msg = string.join(errors, '\n')
            if REQUEST:
                REQUEST.set('manage_tabs_message', msg)
                REQUEST.set('management_view', 'Schedule')
                return self.manage_scheduleForm(self, REQUEST)
                            
            raise AssertionError, msg
        
        self.minute = minute
        self.hour = hour
        self.month = month
        self.day_of_month = day_of_month
        self.day_of_week = day_of_week
        self.tz = tz
        # time is actually calculable but we need it as an attribute to actually
        # index ourselves by it in the catalog ;)
        self.time = self.nextEventTime()
        self._active = active
        #
        # go place it in the schedule ...
        #
        if active:
            self.manage_schedule()
        else:
            self.manage_unschedule()

        if REQUEST:
            REQUEST.set('management_view', 'Schedule')
            return self.manage_scheduleForm(self,REQUEST)

    def manage_unschedule(self, REQUEST=None):
        """
        notify the scheduler
        """
        if self._active:
            self._active = False
            self.getPhysicalRoot().ZSchedulerTool.unschedule(self)
        elif REQUEST:
            REQUEST.set('manage_tabs_message', 'Not active!')
        if REQUEST:
            REQUEST.set('management_view', 'Schedule')
            return self.manage_scheduleForm(self, REQUEST)
            
    def manage_schedule(self, REQUEST=None):
        """
        notify the scheduler
        """
        try:
	    self.getPhysicalRoot().ZSchedulerTool.unschedule(self)
        except:
            pass
        self.getPhysicalRoot().ZSchedulerTool.schedule(self)
        if REQUEST:
            REQUEST.set('management_view', 'Schedule')
            return self.manage_scheduleForm(self, REQUEST)
            
    def test_schedule(self, start_time=None, size=10):
        """
        visually display next scheduled times ...
        """
        if start_time is None:
            start_time = DateTime(self.tz)

        if start_time.timezone() != self.tz:
            start_time = start_time.toZone(self.tz)

        results = []
        for index in range(0, size):
	    try:
                start_time = self.nextEventTime(start_time)
	    except:
		break
            results.append(start_time)
        return results
    
    def manage_repair(self, REQUEST=None):
        """ hmmm fix conversion of derived classes ..."""
        if not getattr(aq_base(self), 'callable_id', None):
            self.callable_id = self.id
        if not getattr(aq_base(self), 'log_batch_size', None):
            self.log_batch_size = 15
        if not getattr(aq_base(self), '_logger', None):
            self._logger = Logger()
        LogSupport.manage_repair(self)
        if REQUEST:
            REQUEST.set('management_view', 'Schedule')
            return self.manage_scheduleForm(self, REQUEST)

    def manage_invokeEvent(self, *args, **kw):
        """
        dispatch to callable and log any results
        """
        now = DateTime(self.tz)
        output = self.event()(*args, **kw)
        self.log(self, 'executed...', output)
        self.last_executed = now
        self.time = self.nextEventTime(now)
        self.getPhysicalRoot().ZSchedulerTool.log(self, 'executed...', output)
        return output
        
    #
    # hmm productions go here ...
    #
    # each production modifies it's time parameter and returns a boolean saying
    # if it overflowed ie you should increment it's adjacent time component
    #

    def _matches(self, time):
	"""
	validity check the given DateTime against our spec
	"""
        day_ok = 1

        y = time.year()
        month_len = _MONTH_LEN[(y%4==0 and (y%100!=0 or y%400==0))][time.month()]

        if self.day_of_week != '*':
            day_ok = get_time_for_spec(self.day_of_week, time.dow(), 0, 7) == (time.dow(), 0)
            if self.day_of_month != '*' and not day_ok:            
                day_ok = get_time_for_spec(self.day_of_month, time.day(), 1, month_len) == (time.day(), 0)
        elif self.day_of_month != '*':
            day_ok = get_time_for_spec(self.day_of_month, time.day(), 1, month_len) == (time.day(), 0)

        return get_time_for_spec(self.minute, time.minute(), 0, 59) == (time.minute(), 0) and \
               get_time_for_spec(self.hour, time.hour(), 0, 23) == (time.hour(), 0) and \
               day_ok and get_time_for_spec(self.month, time.month(), 1, 12)

    def _doMinute(self, time):
        hour = int(time.hour())
        new_minute,increment_hr = get_time_for_spec(self.minute, time.minute(), 0, 59)
        if increment_hr:
            hour += 1
            if hour == 24:
                return  self._doMinute(DateTime('%s/%s/%s 00:%0.2i %s' % (
                    time.year(), time.mm(), time.dd(), new_minute, self.tz ) ) + 1)
        return DateTime('%s/%s/%s %0.2i:%0.2i %s' % (
            time.year(), time.mm(), time.dd(), hour, new_minute, self.tz ) )
                               

    def _doHour(self, time):
        hour = int(time.hour())
        day = time.day()
        new_hour,increment_day = get_time_for_spec(self.hour, hour, 0, 23)

        time = DateTime('%s/%s/%s %0.2i:%s %s' % (
            time.year(), time.mm(), time.dd(), new_hour, time.minute(), self.tz
            ) )
        if increment_day:
            time += 1
        return time

    def _doDay(self, time):
        dom_time = dow_time = None
        if self.day_of_month != '*':
            dom_time = self._doDOM(time)

        if self.day_of_week != '*':
            dow_time = self._doDOW(time)

        if dom_time and dow_time:
            return min(dom_time, dow_time)

        return dom_time or dow_time or time
    
    def _doDOW(self, time):
        for x in range(0, 7):
            # if we didn't cycle thru the spec and still bound to ourself, then it's valid ...
            if get_time_for_spec(self.day_of_week, time.dow(), 0, 7) == (time.dow(), 0):
                return time
            time += 1 # increment by one day
        # eek should never get here ...
        raise AssertionError, "invalid day of week for %s (%s)" % (time, self.day_of_week) 

    def _doDOM(self, time):
        # dynamically calculate length of month
        y = time.year()
        month_len = _MONTH_LEN[(y%4==0 and (y%100!=0 or y%400==0))][time.month()]

        new_day,increment_mm = get_time_for_spec(self.day_of_month, time.day(), 1, month_len)

        if increment_mm:
            year = time.year()
            month = time.month()
            if month == 12:
                month = 1
                year += 1
            else:
                month += 1
            if self.day_of_month == '*':
                time = DateTime('%i/%0.2i/01 %0.2i:%0.2i %s' % (
                    year, month, time.hour(), time.minute(), self.tz
                    ) )
                return time

            time = DateTime('%i/%0.2i/%0.2i %0.2i:%0.2i %s' % (
                year, month, int(get_first_time_for_spec(self.day_of_month, 1)),
                time.hour(), time.minute(), self.tz
                ) )
            return time

        time = DateTime('%s/%0.2i/%0.2i %0.2i:%0.2i %s' % (
            time.year(), time.month(), new_day, time.hour(), time.minute(), self.tz ) )
        return time

    def _doMonth(self, time):
        year = time.year()
        new_month,increment_yr = get_time_for_spec(self.month, time.month(), 1, 12)
        if increment_yr: year+=1
        if new_month != time.month():
            dd = 1
        else:
            dd = time.dd()
        time = DateTime('%s/%s/%s %0.2i:%0.2i %s' % (
            year, new_month, dd, time.hour(), time.minute(), self.tz ) )
        return time

    def cronTab(self):
        """
        crontab-formatted times
        """
        return ' '.join((self.minute, 
                         self.hour, 
                         self.day_of_month, 
                         self.month, 
                         self.day_of_week))

    def expandedCronTab(self):
        """
        cron time with ranges etc expanded into commas
        """
        return ' '.join((expand_spec(self.minute), 
                         expand_spec(self.hour), 
                         expand_spec(self.day_of_month), 
                         expand_spec(self.month),
                         expand_spec(self.day_of_week)))
        
AccessControl.class_init.InitializeClass( ZScheduleEvent )


manage_addZScheduleEventForm = PageTemplateFile('zpt/add_event', globals())
def manage_addZScheduleEvent(self, id, callable, title='', REQUEST=None):
    """
    Add an event to the schedule - you *MUST* actually edit the
    ZScheduleEvent to add it to the ZScheduler!
    """

    event = ZScheduleEvent(id, title or '-> %s' % callable, callable)

    self._setObject(id, event)

    if REQUEST:
        REQUEST.RESPONSE.redirect('%s/%s/manage_main' % (REQUEST['URL3'], id))




#
#    Copyright (C) 2004-2011  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import AccessControl, transaction
from Products.ZScheduler.interfaces.ITimer import ITimer
from threading import Thread
from DateTime import DateTime
from Acquisition import aq_base
from Products.ZScheduler.timers.Base import ThreadedTimer

# need a handle to the thread ...
worker = None

class ZopeSingleThreaded(ThreadedTimer):
    """
    A single-threaded timer which goes to sleep, waking to fire the next event in the queue
    """        
    meta_type = 'ZopeSingleThreaded'

    def _load(self):
        while 1:
            event = scheduler.nextEvent()
            while event:
                now = DateTime()
                e_time = event.time.timeTime()
                interval = e_time - now.timeTime()
            
                scheduler.semaphore.wait(interval)
                
                now = DateTime()
                if e_time - now.timeTime() <= 0:
                    self._dispatch(event)

    def _dispatch(self, event):
        """
        overrideable dispatching driver
        """
        event.manage_invokeEvent()
        transaction.get().commit()

AccessControl.class_init.InitializeClass(ZopeSingleThreaded)



#
#    Copyright (C) 2004-2012  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import AccessControl, Products, DateTime, threading, os, transaction, urlparse
from Acquisition import aq_base
from AccessControl.Permissions import view_management_screens, change_configuration, access_contents_information
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from ZQueue import ZQueue
from Log import Logger, LogSupport
from timers import timers
from interfaces import IScheduler
from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse
from ZPublisher.BaseRequest import RequestContainer

from zope.interface import implements

devnull = open('/dev/null', 'r+')

class ZScheduler(PropertyManager, LogSupport):
    """
    An event scheduler

    This class provides the ZMI wrapper which delegates to it's two
    internal structures, a queue containing the events, and a timer
    which dispatches them.
    """
    meta_type = 'ZScheduler'

    implements(IScheduler)

    icon='misc_/ZScheduler/scheduler'
    id = 'ZSchedulerTool'
    title = 'Event Scheduler'

    # this guys lives in memory not ZODB ...
    semaphore = threading.Event()
    
    property_extensible_schema__ = 0
    _properties = (
        {'id':'timezone',  'type':'selection',          'mode':'w',
         'select_variable': 'timezones' },
        {'id':'callables', 'type':'multiple selection', 'mode':'w',
         'select_variable': 'products'},
        {'id':'timer_type','type':'selection',          'mode':'w',
         'select_variable': 'timers'},
        {'id':'queue_batch_size', 'type':'int', 'mode':'w'},
        ) + LogSupport._properties

    __ac_permissions__ = (
        (access_contents_information, ('isActive', 'isScheduled', 'nextEvent')),
        (view_management_screens, ('manage_timer', 'manage_log', 'manage_catalog',
                                   'timezones', 'products', 'timers', 'logger', 'searchResults', 'events')),
        (change_configuration, ('manage_delete', 'manage_unschedule', 'manage_reload', 'manage_restartTimer',
                                'manage_refresh', 'manage_schedule', 'schedule', 'unschedule')),
        ) + PropertyManager.__ac_permissions__ + LogSupport.__ac_permissions__
    
    manage_options = (
        {'label':'Queue',      'action':'manage_main',
         'help':('ZScheduler', 'queue.stx')},
        {'label':'Catalog',    'action':'manage_catalog'},
        {'label':'Timer',      'action':'manage_timer',
         'help':('ZScheduler', 'timer.stx')},
        {'label':'Properties', 'action':'manage_propertiesForm',
         'help':('ZScheduler', 'props.stx')}, ) + LogSupport.manage_options

    manage_main  = PageTemplateFile('zpt/queue', globals())

    def manage_catalog(self, REQUEST):
        """ """
        REQUEST.RESPONSE.redirect('queue/manage_main')

    def manage_timer(self, REQUEST):
        """ """
        REQUEST.RESPONSE.redirect('timer/manage_main')

    def __init__(self):
        LogSupport.__init__(self)
        self.callables = ['Script (Python)', 'DTML Method']
        self.timezone = 'UTC'
        self.timer_type = ''
        self.timer = None
        self.queue_batch_size = 15
        self.log_batch_size = 15
        self.queue = ZQueue('queue')

    def manage_afterAdd(self, item, container):
        addZScheduler(self, None)

    def timezones(self):
        """ available timezones """
        return DateTime.Timezones()

    def products(self):
        """ available callable products """
        results = filter( lambda x: x not in ['ZScheduleEvent'],
                          map( lambda x: x['name'], Products.meta_types ) )
        results.sort()
        return results

    def timers(self):
        return timers()

    def schedule(self, event):
        """
        add/reschedule this event
        """
        event._active = True
        try:
            self.queue.pop(event)
        except:
            pass
        self.queue.push(event)
        event.time = event.nextEventTime()
        transaction.get().savepoint(optimistic=True)
        self.semaphore.set()

    def unschedule(self, event):
        """
        remove event from schedule
        """
        event._active = False
        try:
            self.queue.pop(event)
            transaction.get().savepoint(optimistic=True)
            self.semaphore.set()
        except Exception, e:
            pass
        
    def isScheduled(self, event):
        """
        returns whether or not the event is in our queue
        """
        brainz = self.queue.searchResults(getPhysicalPath='/'.join(event.getPhysicalPath()),
                                          active=True)
        return len(brainz) == 1

    def manage_delete(self, urls, REQUEST=None):
        """
        Remove item(s) from ZODB - well just the catalog actually, as
        we keep on zapping ourselves with this!
        """
        for url in urls:
	    try:
                event = self.queue.get(url)
                event.aq_parent._delObject(event.getId())
	    except:
		self.queue.delete(url)

        if REQUEST:
            return self.manage_main(self, REQUEST)

    def manage_unschedule(self, urls, REQUEST=None):
        """
        reindex (causing status to be false)
        """
	root = self.queue.getPhysicalRoot()
	for url in urls:
	    try:
		event = root.unrestrictedTraverse(url)
	    except:
		self.queue.delete(url)
		continue
	    self.unschedule(url)
		
        transaction.get().savepoint(optimistic=True)
        # tell timer to reset
        self.semaphore.set()
        if REQUEST:
            return self.manage_main(self, REQUEST)
            
    def manage_reload(self, REQUEST=None):
        """
        Reload queue by recreating it from a ZScheduleEvents in the ZODB
        """
        # find events
        self.queue.reload()
        transaction.get().savepoint(optimistic=True)
        # tell timer to reset
        self.semaphore.set()
        if REQUEST:
            return self.manage_main(self, REQUEST)

    def manage_refresh(self, REQUEST={}):
        """
        Refresh queue by removing anything that no longer exists
        """
	root = self.queue.getPhysicalRoot()
	for url in map(lambda x: x.getPath(), self.queue(REQUEST)):
	    try:
		root.unrestrictedTraverse(url).manage_schedule()
	    except:
		self.queue.delete(url)
        if REQUEST:
	    return self.manage_main(self, REQUEST)

    def manage_schedule(self, urls, REQUEST=None):
        """
        Reschedule selected items
        """
	root = self.queue.getPhysicalRoot()
	for url in urls:
	    try:
		event = root.unrestrictedTraverse(url)
	    except:
		self.queue.delete(url)
		continue
	    self.schedule(event)
		
        transaction.get().savepoint(optimistic=True)
        # tell timer to reset
        self.semaphore.set()
        if REQUEST:
            return self.manage_main(self, REQUEST)

    def _updateProperty(self, id, value):
        """
        can't be arsed painting a custom properties page, but we need
        to dynamically change our timer implementation ...
        """
        if id == 'timer_type':
            if self.timer_type != value:
                self.timer and self.timer._stop(self)
                exec """
from timers.%s.%s import %s
self.timer = %s()
""" % (value, value, value, value)
                self.timer._start(self)
        PropertyManager._updateProperty(self, id, value)

    def searchResults(self, REQUEST={}, **kw):
        """ query the queue/catalog, returning brains """
        query = dict(REQUEST)
        if kw:
            query.update(kw)

        if not query.has_key('active'):
            query['active'] = True

        if not query.has_key('sort_on'):
            query['sort_on'] = 'time'
        if not query.has_key('sort_order'):
            query['sort_order'] = 'ascending'

        return self.queue.searchResults(REQUEST=query)

    def events(self, REQUEST={}, **kw):
        """
        return the *active* events in the queue

        We include request objects with these also so they work
        properly in *normal* Zope contexts as per our thread invokations

        You can pass 'time' queries if you wish ...
        """
        events = []

        proto, domain, directory, parameters, query, fragment = urlparse.urlparse(self.absolute_url())
        if domain.find(':') != -1:
            server_name, server_port = domain.split(':')
        elif proto == 'http':
            server_name, server_port = domain, '80'
        elif proto == 'https':
            server_name, server_port = domain, '443'
        else:
            # hmmm - should we worry about ftp ...
            raise ValueError, proto

        resp = HTTPResponse(stdout=devnull)
        os.environ['SERVER_NAME']    = server_name
        os.environ['SERVER_PORT']    = server_port
        os.environ['SERVER_URL']     = '%s://%s:%s' % (proto, server_name, server_port)
        os.environ['REQUEST_METHOD'] = 'GET'
        request = HTTPRequest(devnull, os.environ, resp)

        for brainz in self.searchResults(REQUEST, **kw):
            try:
                events.append(brainz.getObject())
            except:
                continue

        return map(lambda x,y=request: x.__of__(RequestContainer(REQUEST=y)), events)

    def nextEvent(self):
        return self._queue.getNext()

    def _start(self):
        """ kick off the timer """
        self.timer._start(self)

    def _stop(self):
        """ clean up on exit """
        self.timer._stop(self)

    def isActive(self):
        """ determine if timer is actually running """
        return self.timer.isActive()

    def manage_restartTimer(self, REQUEST=None):
        """
        kick the timer in the guts
        """
        self.timer.manage_restart(self)
        if REQUEST:
            REQUEST.set('manage_tabs_message', 'Restarted timer')
            return self.manage_main(self, REQUEST)

    def manage_repair(self, REQUEST=None):
	"""
	fix catalog(s)
	"""
        if getattr(aq_base(self), '_queue', None):
            queue = aq_base(self._queue)
            delattr(self, '_queue')
            self.queue = queue

        self.queue._repair()

        LogSupport.manage_repair(self)

        # tell timer to reset
        self.semaphore.set()

	if REQUEST:
	    REQUEST.set('manage_tabs_message', 'Repaired')
	    return self.manage_main(self, REQUEST)

AccessControl.class_init.InitializeClass(ZScheduler)


def addZScheduleEvent(ob, event):
    zt = ob.ZSchedulerTool
    zt.queue.push(ob)

def delZScheduleEvent(ob, event):
    # seems some tear-down stuff removes the tool ...
    zt = getattr(ob, 'ZSchedulerTool', None)
    if zt:
        zt.unschedule(ob)
        zt.queue.delete('/'.join(ob.getPhysicalPath()))

def addZScheduler(ob, event):
    # create our timer
    ob._updateProperty('timer_type', 'Dummy')
    # and load it ...
    ob.semaphore.set()




#
#    Copyright (C) 2004-2011  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


from Testing.ZopeTestCase import ZopeTestCase

class TestCrontab(ZopeTestCase):

    def testImport(self):
        # check that it loads OK - we don't get to see normal import errors :(
        from Products.ZScheduler.timers.Crontab import Crontab


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestCrontab))
    return suite



#    Copyright (C) 2004-2009  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import AccessControl, string
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Products.BTreeFolder2.BTreeFolder2 import BTreeFolder2
from Products.ZCatalog.ZCatalog import ZCatalog
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from AccessControl import ClassSecurityInfo
from AccessControl.Permissions import view_management_screens, \
    access_contents_information, delete_objects
from Acquisition import aq_base
from DateTime import DateTime

class LogRec(PropertyManager, SimpleItem):
    """
    A simple log record class
    """
    meta_type = 'Log Record'
    property_extensible_schema__ = 0
    icon='misc_/ZScheduler/logrec'

    _properties = (
        {'id':'timestamp', 'type':'date',   'mode':'r'},
        {'id':'process',   'type':'string', 'mode':'r'},
        {'id':'url',       'type':'string', 'mode':'r'},
        {'id':'title',     'type':'string', 'mode':'r'},
        {'id':'message',   'type':'text',   'mode':'r'},
        )

    manage_options = PropertyManager.manage_options + SimpleItem.manage_options
    
    def __init__(self, id, title, timestamp, process, url, message):
        self.id = id
        self.title = title
        self.timestamp = timestamp
        self.url = url
        self.process = process
        self.message = message

    def event(self):
        """
        return underlying event or None if it no longer exists ...
        """
        try:
            return self.getPhysicalRoot().unrestrictedTraverse(self.url).event()
        except:
            return None

    def manage_afterAdd(self, item, container):
        self.aq_parent.catalog_object(self, '/'.join(self.getPhysicalPath()))

    def manage_beforeDelete(self, item, container):
        self.aq_parent.uncatalog_object('/'.join(self.getPhysicalPath()))

AccessControl.class_init.InitializeClass(LogRec)


class Logger(BTreeFolder2, ZCatalog):
    """
    A simple searcheable log message container

    We haven't yet bothered to implement a search API
    """
    meta_type = 'Logger'
    icon='misc_/ZScheduler/logger'

    id = '_logger'

    def __init__(self, id='_logger', title='ZSchedule Log'):
        BTreeFolder2.__init__(self, self.id)
        ZCatalog.__init__(self, self.id)
        self.title = title
        self.addIndex('timestamp', 'DateIndex')

    def all_meta_types(self):
        return [ ]

    def log(self, timestamp, pid, url, title, msg):
        id = self.generateId()
        self._setObject(id, LogRec(id, title, timestamp, pid, url, msg))

    __call__ = ZCatalog.__call__
    
AccessControl.class_init.InitializeClass(Logger)

class LogSupport(SimpleItem):
    """
    ZMI support for logging
    """
    _security = ClassSecurityInfo()

    __ac_permissions__ = (
        (delete_objects, ('manage_clearLog', 'manage_delLogEntries')),
        (view_management_screens, ('manage_log',)),
        (access_contents_information, ('logger',)),
        ) + SimpleItem.__ac_permissions__

    _properties = (
        {'id':'log_batch_size', 'type':'int', 'mode':'w'},
        )
    
    manage_options = (
        {'label':'Log', 'action':'manage_log', 'help':('ZScheduler', 'log.stx')},
        ) + SimpleItem.manage_options

    manage_log   = PageTemplateFile('zpt/log', globals())

    def __init__(self):
        self._logger = Logger()
        self.log_batch_size = 15

    _security.declarePrivate('log')
    def log(self, event, title, msg=''):
        """
        helper log method
        """
        # if we don't want local logging, we remove the logger, so we
        # just verify it's present here ...
        if getattr(aq_base(self), '_logger', None) is not None:
            self._logger.log(DateTime(), event.getId(), '/%s' % event.absolute_url(1), title, msg)

    def logger(self):
        """ query the log """
        results = []
        query = {'sort_on':'timestamp', 'sort_order':'descending'}
        for brain in self._logger.searchResults(**query):
            # we've 'private' _logger in the path ...
            id = brain.getPath().split('/')[-1]
            results.append(self._logger._getOb(id))
        return results

    def loggerValues(self):
        """
        """
        return self._logger.objectValues()

    def manage_clearLog(self, REQUEST=None):
        """
        remove all log entries
        """
        # f**k knows why the map doesn't clear down everything ...
        while self._logger.objectIds():
            map(lambda x,y=self._logger: y._delObject(x), self._logger.objectIds())
        if REQUEST:
            REQUEST.set('manage_tabs_message', 'Log cleared')
            REQUEST.set('management_view', 'Log')
            return self.manage_log(self, REQUEST)

    def manage_delLogEntries(self, ids, REQUEST=None):
        """
        remove selected log entries
        """
        self._logger.manage_delObjects(ids)
        if REQUEST:
            REQUEST.set('management_view', 'Log')
            return self.manage_log(self, REQUEST)
            
    def manage_repair(self, REQUEST=None):
	"""
	fix catalog(s)
	"""
        if getattr(aq_base(self), '_logger', None):
            # throw it away and start again ...
            self._logger = Logger()
        if REQUEST:
	    REQUEST.set('manage_tabs_message', 'Repaired')
            return self.manage_main(self, REQUEST)

AccessControl.class_init.InitializeClass(LogSupport)



#    Copyright (C) 2005-2008  Corporation of Balclutha. All rights Reserved.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import AccessControl
from ZScheduleEvent import ZScheduleEvent
from Products.CMFCore.FSObject import FSObject
#from Products.CMFCore.utils import expandpath
from Products.CMFCore.DirectoryView import registerFileExtension, registerMetaType
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

class FSZScheduleEvent(FSObject, ZScheduleEvent ):
    """
    A File system ZScheduleEvent
    """
    meta_type = 'FSZScheduleEvent'

    manage_options =  (
        {'label':'Customize', 'action':'manage_main'},
    ) + ZScheduleEvent.manage_options[1:-2]

    __ac_permissions__ = ZScheduleEvent.__ac_permissions__ + FSObject.__ac_permissions__
    manage_main = PageTemplateFile('zpt/fsevent', globals())

    def _createZODBClone(self):
        obj = ZScheduleEvent(self.getId(), self.title, self.callable, self.minute, 
                             self.hour, self.month, self.day_of_month, self.day_of_week, 
                             self.tz)
        return obj

    def __call__(self, *args, **kw):
        '''Calls the script.'''
        self._updateFromFS()
        return ZScheduleEvent.__call__(self, *args, **kw)

    def _readFile(self, reparse):

        """Read the data from the filesystem.

        Read the file (indicated by exandpath(self._filepath), and parse the
        data if necessary.
        """

        #fp = expandpath(self._filepath)

        file = open(self._filepath, 'r')    # not 'rb', as this is a text file!
        try:
            lines = file.readlines()
        finally:
            file.close()

	tz = 'UTC'
        lino=0
        for line in lines:

            lino = lino + 1
            line = line.strip()

            if not line or line[0] == '#':
                continue

            try:
		if line.startswith('title='):
		    self.title = line[6:]
		elif line.startswith('tz='):
		    tz = line[3:]
		else:
		    # we don't want to schedule this yet ...
		    self.status = 0
		    minute,hour,month,day_of_month,day_of_week,callable_id = line.split(' ')
		    self.manage_editSchedule(tz, minute,hour,month,day_of_month,day_of_week, 0) 
		    self.callable_id = callable_id
	    except:
                raise ValueError, ( 'Error processing line %s of %s:\n%s' % (lino,fp,line) )

AccessControl.class_init.InitializeClass(FSZScheduleEvent)


registerFileExtension('sched', FSZScheduleEvent)
registerMetaType('ZScheduleEvent Object', FSZScheduleEvent)



#    Copyright (C) 2005-2011  Corporation of Balclutha. All rights Reserved.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from StringIO import StringIO
from Products.CMFCore.utils import getToolByName
from Products.ZScheduler.config import *


def install(portal):                                       
    out = StringIO()

    setup_tool = getToolByName(portal, 'portal_setup')

    setup_tool.runAllImportStepsFromProfile(
        "profile-Products.ZScheduler:default",
        purge_old=False)

    print >> out, "Successfully installed %s." % PROJECTNAME
    return out.getvalue()

def uninstall(portal, reinstall=False):
    out = StringIO()

    if not reinstall:
        setup_tool = getToolByName(portal, 'portal_setup')
        setup_tool.runAllImportStepsFromProfile('profile-Products.ZScheduler:default')

    print >> out, "Successfully uninstalled %s." % PROJECTNAME
    return out.getvalue()




#
#    Copyright (C) 2004-2012  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import Globals, zLOG
from AccessControl.Permissions import change_configuration
from AccessControl import SecurityManagement, User, class_init
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Products.ZScheduler.interfaces.ITimer import ITimer

from threading import Thread, Event

from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse
from ZPublisher.BaseRequest import RequestContainer

# need a handle to the thread ...
worker = None


class BaseTimer(PropertyManager, SimpleItem):
    """
    Basic timer implementation (ZMI + start/stop)
    """
    __implements__ = (ITimer,)

    id = 'timer'

    _stop_event = Event()

    icon = 'misc_/ZScheduler/timer'

    _properties = (
          {'id':'is_active', 'mode':'w', 'type':'boolean'},
        )

    manage_options = PropertyManager.manage_options + (
        {'label':'Restart', 'action':'manage_restart' },
        ) + SimpleItem.manage_options
    manage_main = PropertyManager.manage_propertiesForm
    property_extensible_schema__ = 0

    __ac_permissions__ = PropertyManager.__ac_permissions__ + (
        (change_configuration, ('manage_restart',)),
       ) + SimpleItem.__ac_permissions__

    def __init__(self):
        self.is_active = False

    def title(self):
        return self.meta_type

    def _start(self, scheduler=None):
        """
        start timer

        essentially just clears the stop semaphore causing the main loop to run
        """
        self._stop_event.clear()

    def _stop(self, scheduler=None):
        """
        stop timer

        essentially just flags the stop semaphore and the main loop should drop out
        """
	self._stop_event.set()
        scheduler = scheduler or self.aq_parent
        scheduler.semaphore.set()

    def isActive(self):
        """
        hard-core function testing underlying timer mechanism's status
        """
        return False

    def _load(self):
        """
        load underlying timer
        """
        pass

    def _unload(self):
        """
        unload underlying timer
        """
        pass


    def manage_restart(self, scheduler=None, REQUEST=None):
        """
        stop/start the timer
        """
        self._stop(scheduler or self.aq_parent)
        self._start(scheduler or self.aq_parent)
        if REQUEST:
            REQUEST.set('manage_tabs_message', 'Restarted')
            return self.manage_main(self, REQUEST)

class_init.InitializeClass(BaseTimer)


class ThreadedTimer(BaseTimer):
    """
    A timer that manages load/reload in a separate thread

    Note that Zope connections do not play well with threads - the connection
    is in another thread.  However, since we're not doing any writes/commits
    in our worker thread, this doesn't affect us.

    Refer to http://www.zopelabs.com/cookbook/1058719806 for a discussion
    about Zope Threads.
    """
    def _start(self, scheduler=None):
	"""
	kick off a listener to refresh crontab's as event's change or are added
	and removed
	"""
        zLOG.LOG('ZScheduler.timer.%s' % self.meta_type, zLOG.INFO, 'starting...')
        if not self.is_active:
            return

        # at app server startup, we don't have a REQUEST - needed to render ZPT
        if not getattr(self, 'REQUEST', None):
            self = self._getContext(self)

        scheduler = scheduler or self.aq_parent

        BaseTimer._start(self, scheduler)

        worker = Thread(None, self._run)
        worker.setName('ZScheduler')
        worker.setDaemon(1)          # detach thread ...
        worker.start()

        # force the run() function to invoke - ie load ... 
        # this is is essentially non-blocking
        scheduler.semaphore.set()

    def _run(self):
        """
        this is our reload procedure
        """
        zLOG.LOG('ZScheduler.timer.%s' % self.meta_type, zLOG.INFO, 'starting listener')
        # ensure system user priviledges
        SecurityManagement.newSecurityManager(None, User.system)

        while 1:
            if self._stop_event.isSet(): 
                zLOG.LOG('ZScheduler.timer.%s' % self.meta_type, zLOG.INFO, 'stopping listener')
                return

            # we need to reopen the connection each time to access the committed changes ...

            # Set up the "application" object that automagically opens connections
            conn = Globals.DB.open()
            app = self._getContext(conn.root()['Application'])
            scheduler = app.ZSchedulerTool

            scheduler.timer._unload()
            scheduler.timer._load()

            scheduler.semaphore.wait()
            # we were interrupted ...
            if scheduler.semaphore.isSet():
                scheduler.semaphore.clear()

            conn.close()
            
    def _stop(self, scheduler):
	"""
	remove crontab
	"""
        zLOG.LOG('ZScheduler.timer.%s' % self.meta_type, zLOG.INFO, 'stopping...')
        BaseTimer._stop(self, scheduler)

        if not self.is_active:
            return
        
        self._unload()

    def _getContext(self, app):
        """
        set up a Zope context
        """
        resp = HTTPResponse(stdout=None)

        from asyncore import socket_map
        http = filter(lambda x:x.__class__.__name__ == 'zhttp_server', socket_map.values())[0]

        env = {
            'SERVER_NAME': http.server_name,
            'SERVER_PORT': str(http.server_port),
            'REQUEST_METHOD':'GET'
            }
        req = HTTPRequest(None, env, resp)
        return app.__of__(RequestContainer(REQUEST = req))

class_init.InitializeClass(ThreadedTimer)



#
# Copyright 2011 Corporation of Balclutha (http://www.balclutha.org)
# 
#                All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software
# and its documentation for any purpose whatsoever is strictly
# prohibited without prior written consent of Corporation of Balclutha.
# 
#
# Corporation of Balclutha DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS, IN NO EVENT SHALL Corporation of Balclutha BE LIABLE FOR
# ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE. 
#

from Products.BastionBase.tests.ZPTTestCase import SkinsTestCase


class TestSkins(SkinsTestCase):
    # Note: This looks like a unit test but isn't
    skinname = 'zscheduler'
    productname = 'ZScheduler'


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestSkins))
    return suite
    



#    Copyright (C) 2009-2010  Corporation of Balclutha. All rights Reserved.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
#    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
#    GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from setuptools import setup, find_packages
from lxml import etree
import os

def read(name):
    try:
        return open(name).read()
    except:
        return ''

long_description=(
        read('README.txt')
#        + '\n' +
#        read('CHANGES.txt')
    )



md = os.path.join(os.path.dirname(__file__), 'Products', 'ZScheduler',
                  'profiles', 'default', 'metadata.xml')
version = etree.fromstring(open(md).read(), 
                           etree.XMLParser()).xpath('/metadata/version')[0].text

setup(name='Products.ZScheduler',
      version=version,
      description="Zope/Plone remote timer",
      long_description=long_description,
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Framework :: Plone",
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        ],
      keywords='ZScheduler',
      author='Alan Milligan',
      author_email='alan.milligan@last-bastion.net',
      url='http://www.last-bastion.net/forge/zscheduler',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['Products'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'plone.directives',
          'plone.supermodel',
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



#    Copyright (C) 2004-2009 Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from Products.CMFCore.utils import registerIcon

import os, os.path
directory = os.path.dirname(__file__)

_timers = filter( lambda x: os.path.isdir(os.path.join(directory, x)) and x not in ['.svn', 'CVS'],
                  os.listdir(directory) )

__all__ = _timers

# make a copy ...
def timers() : return list(_timers)

from AccessControl import ModuleSecurityInfo
ModuleSecurityInfo('Products').declarePublic('ZScheduler')
ModuleSecurityInfo('Products.ZScheduler').declarePublic('timers')
ModuleSecurityInfo('Products.ZScheduler.timers').declarePublic('timers')

def initialize(context):
    for provider in timers():
        exec('''registerIcon(%s.%s, 'times/%s/www/%s.gif', globals())''' % (provider, 
                                                                            provider, 
                                                                            provider.lower()))





#
#    Copyright (C) 2004-2015  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


from Testing import ZopeTestCase
from Products.ZScheduler.ZScheduleEvent import *
from DateTime import DateTime

ZopeTestCase.installProduct('ZScheduler')



class TestScheduleEvent(ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        self.event = ZScheduleEvent('test', 'testing', 'no callable')
        self.now = DateTime('Jan 1 2004 00:00:00 UTC')

    def testRangeExpansions(self):
        self.assertEqual(get_time_for_spec('*', 5, 0, 59), (5,0))
        self.assertEqual(get_time_for_spec('5', 5, 0, 59), (5,0))
        self.assertEqual(get_time_for_spec('5', 6, 0, 59), (5,1))
        self.assertEqual(get_time_for_spec('1-5', 5, 0, 59), (5,0))
        self.assertEqual(get_time_for_spec('1,3,5', 5, 0, 59), (5,0))
        self.assertEqual(get_time_for_spec('1,3,5-20', 5, 0, 59), (5,0))
        self.assertEqual(get_time_for_spec('1,3', 5, 1, 12), (1,1))

    def testIntervalExpansion(self):
        self.assertEqual(expand_spec('1-5'), '1,2,3,4,5')

    def testSkipsExpansion(self):
        self.assertEqual(expand_spec('10-30/5'), '10,15,20,25,30')

    def testRangesExpansion(self):
        self.assertEqual(expand_spec('1-2,10-12'), '1,2,10,11,12')

    def testNextDay(self):
        self.assertEqual(next_day(DateTime('2004/01/01'), 'UTC'), DateTime('2004/01/02 UTC'))
        self.assertEqual(next_day(DateTime('2004/01/31'), 'UTC'), DateTime('2004/02/01 UTC'))
        self.assertEqual(next_day(DateTime('2004/12/31'), 'UTC'), DateTime('2005/01/01 UTC'))

    def testNextMonth(self):
        self.assertEqual(next_month(DateTime('2004/01/01'), 'UTC'), DateTime('2004/02/01 UTC'))
        self.assertEqual(next_month(DateTime('2004/01/31'), 'UTC'), DateTime('2004/02/01 UTC'))
        self.assertEqual(next_month(DateTime('2004/12/31'), 'UTC'), DateTime('2005/01/01 UTC'))
        
    def testMinute(self):
        self.event.manage_editSchedule('US/Eastern', '1', '*', '*', '*', '*', 0)
        # note tz difference with now in UTC ...
        self.assertEqual(self.event.test_schedule(self.now, 5),
                         [ DateTime('Dec 31 2003 19:01:00 US/Eastern'),
                           DateTime('Dec 31 2003 20:01:00 US/Eastern'),
                           DateTime('Dec 31 2003 21:01:00 US/Eastern'),
                           DateTime('Dec 31 2003 22:01:00 US/Eastern'),
                           DateTime('Dec 31 2003 23:01:00 US/Eastern') ])
    def testHour(self):
        self.event.manage_editSchedule('US/Eastern', '1', '1-2', '*', '*', '*', 0)
        self.assertEqual(self.event.test_schedule(self.now, 5),
                         [ DateTime('Jan 1 2004 01:01:00 US/Eastern'),
                           DateTime('Jan 1 2004 02:01:00 US/Eastern'),
                           DateTime('Jan 2 2004 01:01:00 US/Eastern'),
                           DateTime('Jan 2 2004 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2004 01:01:00 US/Eastern') ])

    def testMonth(self):
        self.event.manage_editSchedule('US/Eastern', '1', '1', '11,12', '*', '*', 0)
        self.assertEqual(self.event.test_schedule(self.now, 5),
                         [ DateTime('Nov 1 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 2 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 3 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 4 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 5 2004 01:01:00 US/Eastern') ])

    def testDayOfMonth(self):
        self.event.manage_editSchedule('US/Eastern', '1', '1', '11', '3,11', '*', 0)
        self.assertEqual(self.event.test_schedule(self.now, 6),
                         [ DateTime('Nov  3 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 11 2004 01:01:00 US/Eastern'),
                           DateTime('Nov  3 2005 01:01:00 US/Eastern'),
                           DateTime('Nov 11 2005 01:01:00 US/Eastern'),
                           DateTime('Nov  3 2006 01:01:00 US/Eastern'),
                           DateTime('Nov 11 2006 01:01:00 US/Eastern') ])
    def testDayOfWeek(self):
        self.event.manage_editSchedule('US/Eastern', '1', '1', '11', '*', '2', 0)
        self.assertEqual(self.event.test_schedule(self.now, 6),
                         [ DateTime('Nov  2 2004 01:01:00 US/Eastern'),
                           DateTime('Nov  9 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 16 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 23 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 30 2004 01:01:00 US/Eastern'),
                           DateTime('Nov  1 2005 01:01:00 US/Eastern') ])

    def testDOWandMonth(self):
        self.event.manage_editSchedule('US/Eastern', '1', '1', '11', '3,11', '2', 0)
        self.assertEqual(self.event.test_schedule(self.now, 6),
                         [ DateTime('Nov  2 2004 01:01:00 US/Eastern'),
                           DateTime('Nov  3 2004 01:01:00 US/Eastern'),
                           DateTime('Nov  9 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 11 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 16 2004 01:01:00 US/Eastern'),
                           DateTime('Nov 23 2004 01:01:00 US/Eastern'), ])

    def testIntraDay(self):
        self.event.manage_editSchedule('US/Eastern', '1-2,31-32', '1', '*', '*', '*', 0)
        self.assertEqual(self.event.test_schedule(DateTime('Jan 01 2004 1:24 US/Eastern'), 5),
                         [ DateTime('Jan 1 2004 01:31:00 US/Eastern'),
                           DateTime('Jan 1 2004 01:32:00 US/Eastern'),
                           DateTime('Jan 2 2004 01:01:00 US/Eastern'),
                           DateTime('Jan 2 2004 01:02:00 US/Eastern'),
                           DateTime('Jan 2 2004 01:31:00 US/Eastern'), ])
        
    def testMinuteOverflow(self):
        self.event.manage_editSchedule('US/Eastern', '0-2,59', '1,2', '*', '*', '*', 0)
        self.assertEqual(self.event.test_schedule(DateTime('Jan 01 2004 1:24 US/Eastern'), 5),
                         [ DateTime('Jan 1 2004 01:59:00 US/Eastern'),
                           DateTime('Jan 1 2004 02:00:00 US/Eastern'),
                           DateTime('Jan 1 2004 02:01:00 US/Eastern'),
                           DateTime('Jan 1 2004 02:02:00 US/Eastern'),
                           DateTime('Jan 1 2004 02:59:00 US/Eastern'), ])
        #
        # this tests a screwed time rollover at Enchant ...
        #
        self.event.manage_editSchedule('UTC', '1', '1', '*', '*', '*', 0)
        self.assertEqual(self.event.test_schedule(DateTime('Jan 01 2004 23:24 UTC'), 5),
                         [ DateTime('Jan 2 2004 01:01:00 UTC'),
                           DateTime('Jan 3 2004 01:01:00 UTC'),
                           DateTime('Jan 4 2004 01:01:00 UTC'),
                           DateTime('Jan 5 2004 01:01:00 UTC'),
                           DateTime('Jan 6 2004 01:01:00 UTC'), ])

    
    def testHourOverflow(self):
        self.event.manage_editSchedule('US/Eastern', '1', '2', '1', '*', '*', 0)
        self.assertEqual(self.event.test_schedule(DateTime('Jan 02 2004 1:24 US/Eastern'), 5),
                         [ DateTime('Jan 2 2004 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2004 02:01:00 US/Eastern'),
                           DateTime('Jan 4 2004 02:01:00 US/Eastern'),
                           DateTime('Jan 5 2004 02:01:00 US/Eastern'),
                           DateTime('Jan 6 2004 02:01:00 US/Eastern'), ])
    
    def testDayOverflow(self):
        self.event.manage_editSchedule('US/Eastern', '1', '2', '1', '3', '*', 0)
        self.assertEqual(self.event.test_schedule(DateTime('Jan 02 2004 1:24 US/Eastern'), 5),
                         [ DateTime('Jan 3 2004 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2005 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2006 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2007 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2008 02:01:00 US/Eastern'), ])

    def testTimeZone(self):
        # dunno about this - timezone's seem f**ked to me ...
        tm = self.now.toZone('UTC')
        self.event.manage_editSchedule('US/Eastern', '1', '2', '1', '3', '*', 0)
        self.assertEqual(self.event.test_schedule(tm, 5),
                         [ DateTime('Jan 3 2004 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2005 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2006 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2007 02:01:00 US/Eastern'),
                           DateTime('Jan 3 2008 02:01:00 US/Eastern'), ])

    def testRollOverDay(self):
        tm = DateTime('2004/09/12 23:57 UTC')
        self.event.manage_editSchedule('UTC', '0-59', '23,0', '*', '*', '*', 0)
        self.assertEqual(self.event.test_schedule(tm, 5),
                         [ DateTime('2004/09/12 23:58 UTC'),
                           DateTime('2004/09/12 23:59 UTC'),
                           DateTime('2004/09/13 00:00 UTC'),
                           DateTime('2004/09/13 00:01 UTC'),
                           DateTime('2004/09/13 00:02 UTC'),
                           ])

    def testCronStrs(self):
        self.event.manage_editSchedule('UTC', '0-59/10', '23,0', '1-5', '*', '*', 0)
        self.assertEqual(self.event.cronTab(), '0-59/10 23,0 * 1-5 *')
        self.assertEqual(self.event.expandedCronTab(), '0,10,20,30,40,50 23,0 * 1,2,3,4,5 *')

    def testTimezone(self):
        # Timezone may/will move from DateTime ...
        self.failUnless(len(self.event.timezones()) > 20)

    def testZPTRenders(self):
        content = self.event.manage_scheduleForm()
        # yay ...



def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestScheduleEvent))
    return suite



#
#    Copyright (C) 2011-2015  Corporation of Balclutha. All rights Reserved.
#
#    For your open source solution, bureau service and consultancy requirements,
#    visit us at http://www.balclutha.org or http://www.last-bastion.net.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from Testing import ZopeTestCase
from Products.PloneTestCase import PloneTestCase as ptc
from Products.ZScheduler.ZScheduleEvent import ZScheduleEvent

ZopeTestCase.installProduct('ZScheduler')

ptc.setupPloneSite(products=['ZScheduler'])


class TestScheduler(ptc.PloneTestCase):

    def afterSetUp(self):

        portal = self.app
        #portal = self.portal
        portal._setObject('event1', ZScheduleEvent('event1', 'testing', 'no callable'))
        portal._setObject('event2', ZScheduleEvent('event2', 'testing', 'no callable'))
        portal._setObject('event3', ZScheduleEvent('event3', 'testing', 'no callable'))

        self.event1 = portal.event1

    def testSetup(self):
        self.failUnless('ZSchedulerTool' in self.app.objectIds())

    def testQueueCore(self):
        self.assertEqual(self.portal.event1._active, False)
        self.assertEqual(self.portal.event1.active, False)

        catalog = self.app.ZSchedulerTool.queue
        self.assertEqual(len(catalog.searchResults()), 3)
        self.assertEqual(len(catalog.searchResults(active=False)), 3)

    def testSchedulerAPI(self):
        zscheduler = self.app.ZSchedulerTool
        self.assertEqual(len(zscheduler.events()), 0)
        self.assertEqual(len(zscheduler.events(active=True)), 0)
        self.assertEqual(len(zscheduler.events(active=False)), 3)

        self.event1.manage_editSchedule('EST', '5', '*', '*', '*','*', True)
        self.assertEqual(zscheduler.events(active=True), [self.event1])


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestScheduler))
    return suite
